from enum import IntEnum
from typing import List

from pymafia import ash, player
from pymafia.datatypes import Familiar, Item
from pymafia.preference import get_property

ITEM = Item("mumming trunk")


class Costume(IntEnum):
    MEAT = 0
    MP = 1
    MUSCLE = 2
    ITEM = 3
    MYST = 4
    HP = 5
    MOXIE = 6


def have() -> bool:
    """Return True if the player has the mumming trunk available, False otherwise."""
    return player.have(ITEM)


def costumes_used() -> List[Costume]:
    """Return a list of the costumes applied today."""
    return [Costume(int(x)) for x in get_property("_mummeryUses").split(",") if x]


def apply_costume(familiar: Familiar, costume: Costume) -> bool:
    """Dress up a familiar with a costume."""
    if not have():
        return False
    if costume in costumes_used():
        return True  # TODO return True if costume is equipped on desired familiar
    if not player.have(familiar):
        return False

    ash.use_familiar(familiar)
    return ash.cli_execute(f"mummery {int(costume)}")



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.item import Item


@dataclass(frozen=True, order=True)
class Familiar:
    id: int = km.DataTypes.FAMILIAR_INIT.contentLong
    name: str = km.DataTypes.FAMILIAR_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        id = km.FamiliarDatabase.getFamiliarId(key) if isinstance(key, str) else key
        name = km.FamiliarDatabase.getFamiliarName(id)
        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "id", id)
        object.__setattr__(self, "name", name)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Familiar]:
        from pymafia.ash import from_java

        values = km.DataTypes.FAMILIAR_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def hatchling(self) -> Item:
        from pymafia.datatypes.item import Item

        try:
            return Item(km.FamiliarDatabase.getFamiliarLarva(self.id))
        except ValueError:
            return Item()

    @property
    def image(self) -> str:
        return km.FamiliarDatabase.getFamiliarImageLocation(self.id)

    @property
    def nickname(self) -> str:
        fam = km.KoLCharacter.ownedFamiliar(self.name)
        return fam.getName() if fam.isPresent() else ""

    @property
    def owner(self) -> str:
        fam = km.KoLCharacter.ownedFamiliar(self.name)
        return fam.getOwner() if fam.isPresent() else ""

    @property
    def owner_id(self) -> int:
        fam = km.KoLCharacter.ownedFamiliar(self.name)
        return fam.getOwnderId() if fam.isPresent() else 0

    @property
    def experience(self) -> int:
        fam = km.KoLCharacter.usableFamiliar(self.id)
        return fam.getTotalExperience() if fam is not None else 0

    @property
    def charges(self) -> int:
        fam = km.KoLCharacter.usableFamiliar(self.id)
        return fam.getCharges() if fam is not None else 0

    @property
    def drop_name(self) -> str:
        return km.FamiliarData.dropName(self.id) or ""

    @property
    def drop_item(self) -> Item:
        from pymafia.datatypes.item import Item

        item = km.FamiliarData.dropItem(self.id)
        return Item(item.getItemId()) if item is not None else Item()

    @property
    def drops_today(self) -> int:
        return km.FamiliarData.dropsToday(self.id)

    @property
    def drops_limit(self) -> int:
        return km.FamiliarData.dropsToday(self.id)

    @property
    def fights_today(self) -> int:
        return km.FamiliarData.fightsToday(self.id)

    @property
    def fights_limit(self) -> int:
        return km.FamiliarData.fightDailyCap(self.id)

    @property
    def combat(self) -> bool:
        return km.FamiliarDatabase.isCombatType(self.id)

    @property
    def physical_damage(self) -> bool:
        return km.FamiliarDatabase.isCombat0Type(self.id)

    @property
    def elemental_damage(self) -> bool:
        return km.FamiliarDatabase.isCombat1Type(self.id)

    @property
    def block(self) -> bool:
        return km.FamiliarDatabase.isBlockType(self.id)

    @property
    def delevel(self) -> bool:
        return km.FamiliarDatabase.isDelevelType(self.id)

    @property
    def hp_during_combat(self) -> bool:
        return km.FamiliarDatabase.isHp0Type(self.id)

    @property
    def mp_during_combat(self) -> bool:
        return km.FamiliarDatabase.isMp0Type(self.id)

    @property
    def other_action_during_combat(self) -> bool:
        return km.FamiliarDatabase.isOther0Type(self.id)

    @property
    def hp_after_combat(self) -> bool:
        return km.FamiliarDatabase.isHp1Type(self.id)

    @property
    def mp_after_combat(self) -> bool:
        return km.FamiliarDatabase.isMp1Type(self.id)

    @property
    def other_action_after_combat(self) -> bool:
        return km.FamiliarDatabase.isOther1Type(self.id)

    @property
    def passive(self) -> bool:
        return km.FamiliarDatabase.isPassiveType(self.id)

    @property
    def underwater(self) -> bool:
        return km.FamiliarDatabase.isUnderwaterType(self.id)

    @property
    def variable(self) -> bool:
        return km.FamiliarDatabase.isVariableType(self.id)

    @property
    def attributes(self) -> list[str]:
        attrs = km.FamiliarDatabase.getFamiliarAttributes(self.id)
        return list(attrs) if attrs is not None else []

    @property
    def poke_level(self) -> int:
        fam = km.KoLCharacter.usableFamiliar(self.id)
        return fam.getPokeLevel() if fam is not None else 0

    @property
    def poke_level_2_power(self) -> int:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getPower2() if data is not None else 0

    @property
    def poke_level_2_hp(self) -> int:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getHP2() if data is not None else 0

    @property
    def poke_level_3_power(self) -> int:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getPower3() if data is not None else 0

    @property
    def poke_level_3_hp(self) -> int:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getHP3() if data is not None else 0

    @property
    def poke_level_4_power(self) -> int:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getPower4() if data is not None else 0

    @property
    def poke_level_4_hp(self) -> int:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getHP4() if data is not None else 0

    @property
    def poke_move_1(self) -> str:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getMove1() if data is not None else ""

    @property
    def poke_move_2(self) -> str:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getMove2() if data is not None else ""

    @property
    def poke_move_3(self) -> str:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getMove3() if data is not None else ""

    @property
    def poke_attribute(self) -> str:
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return data.getAttribute() if data is not None else ""



from pymafia import ash, player
from pymafia.combat import Macro
from pymafia.datatypes import Location, Monster, Skill
from pymafia.preference import get_property

PASSIVE = Skill("Comprehensive Cartography")
SKILL = Skill("Map the Monsters")


def have() -> bool:
    """Return True if the player has the Comprehensive Cartography skill, False otherwise."""
    return player.have(PASSIVE)


def monsters_mapped() -> int:
    """Return the number of Map the Monsters skill uses today."""
    return get_property("_monstersMapped", int)


def map_monster(location: Location, monster: Monster, macro: Macro = Macro()) -> bool:
    """Map to a monster in a location."""
    if not have():
        return False
    if monsters_mapped() >= 3:
        return False
    if not ash.can_adventure(location):
        return False

    if not get_property("mappingMonsters", bool):
        ash.use_skill(SKILL)

    turns = ash.my_turncount()
    while not player.in_combat():
        if ash.my_turncount() > turns:
            raise RuntimeError("Map the Monsters unsuccessful?")
        ash.visit_url(location.url)
        if player.in_choice(1435):
            ash.run_choice(1, False, f"heyscriptswhatsupwinkwink={monster.id}")
            ash.run_combat(macro)
            return True
        else:
            ash.run_choice(-1, False)
    return False



from enum import IntEnum

from pymafia import ash
from pymafia.datatypes import Location
from pymafia.preference import get_property

LOCATION = Location("The X-32-F Combat Training Snowman")


class Setting(IntEnum):
    MUSCLE = 1
    MYSTICALITY = 2
    MOXIE = 3
    TOURNAMENT = 4


def have() -> bool:
    """Return True if the player has The Snojo available, False otherwise"""
    return get_property("snojoAvailable", bool)


def current_setting() -> Setting:
    """Return the current snojo setting."""
    return Setting[get_property("snojoSetting")]


def free_fights_today() -> int:
    """Return the number of free snojo fights used today."""
    return get_property("_snojoFreeFights", int)


def free_fights_left() -> int:
    """Return the number of free snojo fights left today."""
    return 10 - free_fights_today()


def change_setting(setting: Setting) -> bool:
    """Change the snojo setting."""
    if not have():
        return False
    if current_setting() is setting:
        return True

    ash.visit_url("place.php?whichplace=snojo&action=snojo_controller")
    ash.run_choice(int(setting))
    return True



from pymafia import ash, player
from pymafia.datatypes import Familiar
from pymafia.preference import get_property

FAMILIAR = Familiar("Pair of Stomping Boots")


def have() -> bool:
    """Return True if the player has the Pair of Stomping Boots in their terrarium, False otherwise."""
    return player.have(FAMILIAR)


def runaways_used() -> int:
    """Return the number of free bander runaways used today."""
    return get_property("banderRunaways", int)


def runaways_left() -> int:
    """Return the total number of free bander runaways the player can get from their Stomping Boots."""
    return (ash.familiar_weight(FAMILIAR) + ash.weight_adjustment()) // 5



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from jpype import JClass

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.bounty import Bounty

Integer = JClass("java.lang.Integer")


@dataclass(frozen=True, order=True)
class Location:
    kol_adventure: Any = field(
        default=km.DataTypes.LOCATION_INIT.content, compare=False
    )
    id: int = -1  # LOCATION_INIT does not specify an id
    name: str = km.DataTypes.LOCATION_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        kol_adventure = (
            km.AdventureDatabase.getAdventure(key)
            if isinstance(key, str)
            else km.AdventureDatabase.getAdventureByURL(
                f"adventure.php?snarfblat={key}"
            )
        )
        if kol_adventure is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "kol_adventure", kol_adventure)
        object.__setattr__(self, "id", kol_adventure.getSnarfblat())
        object.__setattr__(self, "name", kol_adventure.getAdventureName())

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Location]:
        from pymafia.ash import from_java

        values = km.DataTypes.LOCATION_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def url(self) -> str:
        return f"adventure.php?snarfblat={self.id}"

    @property
    def nocombats(self) -> bool:
        return self.kol_adventure is not None and self.kol_adventure.isNonCombatsOnly()

    @property
    def combat_percent(self) -> float:
        if self.kol_adventure is None:
            return 0
        area = self.kol_adventure.getAreaSummary()
        return area.areaCombatPercent() if area is not None else 0

    @property
    def zone(self) -> str:
        return self.kol_adventure.getZone() if self.kol_adventure is not None else ""

    @property
    def parent(self) -> str:
        return (
            self.kol_adventure.getParentZone() if self.kol_adventure is not None else ""
        )

    @property
    def parentdesc(self) -> str:
        return (
            self.kol_adventure.getParentZoneDescription()
            if self.kol_adventure is not None
            else ""
        )

    @property
    def environment(self) -> str:
        return (
            self.kol_adventure.getEnvironment()
            if self.kol_adventure is not None
            else ""
        )

    @property
    def root(self) -> str:
        return (
            self.kol_adventure.getRootZone() if self.kol_adventure is not None else ""
        )

    @property
    def bounty(self) -> Bounty:
        from pymafia.datatypes.bounty import Bounty

        if self.kol_adventure is None:
            return Bounty()
        bounty = km.AdventureDatabase.getBounty(self.kol_adventure)
        return Bounty(bounty.getName()) if bounty is not None else Bounty()

    @property
    def combat_queue(self) -> list[str]:
        if self.kol_adventure is None:
            return []

        zone_queue = km.AdventureQueueDatabase.getZoneQueue(self.kol_adventure)
        return list(zone_queue) if zone_queue is not None else []

    @property
    def noncombat_queue(self) -> list[str]:
        if self.kol_adventure is None:
            return []
        zone_queue = km.AdventureQueueDatabase.getZoneNoncombatQueue(self.kol_adventure)
        return list(zone_queue) if zone_queue is not None else []

    @property
    def turns_spent(self) -> int:
        return (
            km.AdventureSpentDatabase.getTurns(self.kol_adventure, True)
            if self.kol_adventure is not None
            else 0
        )

    @property
    def kisses(self) -> int:
        return (
            km.FightRequest.dreadKisses(self.kol_adventure)
            if self.kol_adventure is not None
            else 0
        )

    @property
    def recommended_stat(self) -> int:
        return (
            self.kol_adventure.getRecommendedStat()
            if self.kol_adventure is not None
            else 0
        )

    @property
    def poison(self) -> int:
        if self.kol_adventure is None:
            return Integer.MAX_VALUE
        area = self.kol_adventure.getAreaSummary()
        return area.poison() if area is not None else Integer.MAX_VALUE

    @property
    def water_level(self) -> int:
        if self.kol_adventure is None or not km.KoLCharacter.inRaincore():
            return 0
        return self.kol_adventure.getWaterLevel()

    @property
    def wanderers(self) -> bool:
        return self.kol_adventure is not None and self.kol_adventure.hasWanderers()

    @property
    def fire_level(self) -> int:
        if self.kol_adventure is None or not km.KoLCharacter.inFirecore():
            return 0
        return km.WildfireCampRequest.getFireLevel(self.kol_adventure)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from jpype import JClass

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.effect import Effect
    from pymafia.datatypes.element import Element
    from pymafia.datatypes.phylum import Phylum


Integer = JClass("java.lang.Integer")


@dataclass(frozen=True, order=True)
class Monster:
    monster: Any = field(default=km.DataTypes.MONSTER_INIT.content, compare=False)
    id: int = km.DataTypes.MONSTER_INIT.contentLong
    name: str = km.DataTypes.MONSTER_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        monster = (
            km.MonsterDatabase.findMonster(key, True)
            if isinstance(key, str)
            else km.MonsterDatabase.findMonsterById(key)
        )
        if monster is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "monster", monster)
        object.__setattr__(self, "id", monster.getId())
        object.__setattr__(self, "name", monster.getName())

    def __str__(self) -> str:
        ids = km.MonsterDatabase.getMonsterIds(self.name, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Monster]:
        from pymafia.ash import from_java

        values = km.DataTypes.MONSTER_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def article(self) -> str:
        return self.monster.getArticle() if self.monster is not None else ""

    @property
    def base_hp(self) -> int:
        return self.monster.getHP() if self.monster is not None else 0

    @property
    def base_attack(self) -> int:
        return self.monster.getAttack() if self.monster is not None else 0

    @property
    def raw_hp(self) -> int:
        return self.monster.getRawHP() if self.monster is not None else 0

    @property
    def raw_attack(self) -> int:
        return self.monster.getRawAttack() if self.monster is not None else 0

    @property
    def raw_defense(self) -> int:
        return self.monster.getRawDefense() if self.monster is not None else 0

    @property
    def base_defense(self) -> int:
        return self.monster.getDefense() if self.monster is not None else 0

    @property
    def base_initiative(self) -> int:
        return self.monster.getInitiative() if self.monster is not None else 0

    @property
    def raw_initiative(self) -> int:
        return self.monster.getRawInitiative() if self.monster is not None else 0

    @property
    def attack_element(self) -> Element:
        from pymafia.datatypes.element import Element

        return (
            Element(self.monster.getAttackElement().toString())
            if self.monster is not None
            else Element()
        )

    @property
    def attack_elements(self) -> list[Element]:
        from pymafia.datatypes.element import Element

        if self.monster is None:
            return []
        return [Element(x.toString()) for x in self.monster.getAttackElements()]

    @property
    def defense_element(self) -> Element:
        from pymafia.datatypes.element import Element

        return (
            Element(self.monster.getDefenseElement().toString())
            if self.monster is not None
            else Element()
        )

    @property
    def physical_resistance(self) -> int:
        return self.monster.getPhysicalResistance() if self.monster is not None else 0

    @property
    def elemental_resistance(self) -> int:
        return self.monster.getElementalResistance() if self.monster is not None else 0

    @property
    def min_meat(self) -> int:
        return self.monster.getMinMeat() if self.monster is not None else 0

    @property
    def max_meat(self) -> int:
        return self.monster.getMaxMeat() if self.monster is not None else 0

    @property
    def min_sprinkles(self) -> int:
        return self.monster.getMinSprinkles() if self.monster is not None else 0

    @property
    def max_sprinkles(self) -> int:
        return self.monster.getMaxSprinkles() if self.monster is not None else 0

    @property
    def base_mainstat_exp(self) -> float:
        return self.monster.getExperience() if self.monster is not None else 0

    @property
    def group(self) -> int:
        return self.monster.getGroup() if self.monster is not None else 0

    @property
    def phylum(self) -> Phylum:
        from pymafia.datatypes.phylum import Phylum

        return (
            Phylum(self.monster.getPhylum().toString())
            if self.monster is not None
            else Phylum()
        )

    @property
    def poison(self) -> Effect:
        from pymafia.datatypes.effect import Effect

        if self.monster is None:
            return Effect()
        poison_level = self.monster.getPoison()
        if poison_level == Integer.MAX_VALUE:
            return Effect()
        poison_name = km.EffectDatabase.getEffectName(
            km.EffectDatabase.POISON_ID[poison_level]
        )
        return Effect(poison_name)

    @property
    def boss(self) -> bool:
        return self.monster.isBoss() if self.monster is not None else False

    @property
    def copyable(self) -> bool:
        return self.monster.isNoCopy() if self.monster is not None else False

    @property
    def image(self) -> str:
        return self.monster.getImage() if self.monster is not None else ""

    @property
    def images(self) -> list[str]:
        return list(self.monster.getImages()) if self.monster is not None else []

    @property
    def random_modifiers(self) -> list[str]:
        return (
            list(self.monster.getRandomModifiers()) if self.monster is not None else []
        )

    @property
    def sub_types(self) -> list[str]:
        return list(self.monster.getSubTypes()) if self.monster is not None else []

    @property
    def manuel_name(self) -> str:
        return self.monster.getManuelName() if self.monster is not None else ""

    @property
    def wiki_name(self) -> str:
        return self.monster.getWikiName() if self.monster is not None else ""

    @property
    def attributes(self) -> str:
        return self.monster.getAttributes() if self.monster is not None else ""



from enum import Enum

from pymafia import ash, player
from pymafia.combat import Macro
from pymafia.datatypes import Item, Monster
from pymafia.preference import get_property

ITEM = Item("Witchess Set")


class Piece(Enum):
    PAWN = Monster("Witchess Pawn")
    KNIGHT = Monster("Witchess Knight")
    BISHOP = Monster("Witchess Bishop")
    ROOK = Monster("Witchess Rook")
    OX = Monster("Witchess Ox")
    KING = Monster("Witchess King")
    WITCH = Monster("Witchess Witch")
    QUEEN = Monster("Witchess Queen")


def have() -> bool:
    """Return True if the player has the Witchess Set in their campground, False otherwise."""
    return ITEM in ash.get_campground()


def fights_today() -> int:
    """Return the number of Witchess fights used today."""
    return get_property("_witchessFights", int)


def fights_left() -> int:
    """Return the number of Witchess fights left today."""
    return 5 - fights_today()


def fight(piece: Piece, macro: Macro = Macro()) -> bool:
    """Fight a Witchess piece."""
    if not have():
        return False
    if fights_left() < 1:
        return False

    ash.visit_url("campground.php?action=witchess")
    ash.run_choice(1)
    ash.visit_url(
        f"choice.php?option=1&pwd&whichchoice=1182&piece={piece.value.id}",
        False,
    )
    if not player.in_combat(piece.value):
        return False
    ash.run_combat(macro)
    return True



__all__ = [
    "barrel_god",
    "boxing_daycare",
    "cartography",
    "cosplay_saber",
    "crimbo_shrub",
    "god_lobster",
    "love_tunnel",
    "mumming_trunk",
    "pantogram",
    "protonic_pack",
    "snojo",
    "songboom",
    "stomping_boots",
    "witchess",
]

from pymafia.iotms import (
    barrel_god,
    boxing_daycare,
    cartography,
    cosplay_saber,
    crimbo_shrub,
    god_lobster,
    love_tunnel,
    mumming_trunk,
    pantogram,
    protonic_pack,
    snojo,
    songboom,
    stomping_boots,
    witchess,
)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from pymafia.kolmafia import km


@dataclass(frozen=True, order=True)
class Element:
    element: Any = field(default=km.DataTypes.ELEMENT_INIT.content, compare=False)
    name: str = km.DataTypes.ELEMENT_INIT.contentString

    def __init__(self, key: str | None = None):
        if (
            isinstance(key, str) and key.casefold() == self.name.casefold()
        ) or key is None:
            return

        element = km.MonsterDatabase.stringToElement(key)
        if element == km.MonsterDatabase.Element.NONE:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "element", element)
        object.__setattr__(self, "name", element.toString())

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Element]:
        from pymafia.ash import from_java

        values = km.DataTypes.ELEMENT_TYPE.allValues()
        return from_java(values)

    @property
    def image(self) -> str:
        # No image for Slime or Supercold in Manuel
        if self.element in [
            km.MonsterDatabase.Element.NONE,
            km.MonsterDatabase.Element.SLIME,
            km.MonsterDatabase.Element.SUPERCOLD,
        ]:
            return "circle.gif"
        if self.element == km.MonsterDatabase.Element.COLD:
            return "snowflake.gif"
        if self.element == km.MonsterDatabase.Element.HOT:
            return "fire.gif"
        if self.element == km.MonsterDatabase.Element.SLEAZE:
            return "wink.gif"
        if self.element == km.MonsterDatabase.Element.SPOOKY:
            return "skull.gif"
        if self.element == km.MonsterDatabase.Element.STENCH:
            return "stench.gif"
        return ""



from __future__ import annotations

from dataclasses import dataclass

from pymafia.kolmafia import km


@dataclass(frozen=True, order=True)
class Slot:
    name: str = km.DataTypes.SLOT_INIT.contentString

    def __init__(self, key: str | None = None):
        if (
            isinstance(key, str) and key.casefold() == self.name.casefold()
        ) or key is None:
            return

        slot = km.EquipmentRequest.slotNumber(key)
        if slot == km.Slot.NONE:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "name", slot.name)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Slot]:
        from pymafia.ash import from_java

        values = km.DataTypes.SLOT_TYPE.allValues()
        return from_java(values)



from enum import Enum

from pymafia import ash, player
from pymafia.datatypes import Item
from pymafia.preference import get_property

ITEM = Item("SongBoom™ BoomBox")


class SongboomSong(str, Enum):
    SPOOKY = "Eye of the Giger"
    FOOD = "Food Vibrations"
    DR = "Remainin' Alive"
    DAMAGE = "These Fists Were Made for Punchin'"
    MEAT = "Total Eclipse of Your Meat"
    NONE = ""


def have() -> bool:
    """Return True if the player has the SongBoom™ BoomBox available, False otherwise."""
    return player.have(ITEM)


def current_song() -> SongboomSong:
    """Return the current SongBoom™ Boombox song."""
    return SongboomSong[get_property("boomBoxSong")]


def song_changes_left() -> int:
    """Return the number of SongBoom™ Boombox song changes left today."""
    return get_property("_boomBoxSongsLeft", int)


def set_song(song: SongboomSong) -> bool:
    """Change the SongBoom™ Boombox song."""
    if not have():
        return False
    if current_song() is song:
        return True
    if song_changes_left() < 1:
        return False

    return ash.cli_execute(f"boombox {song.value}")


def drop_progress() -> int:
    """Return the progress to next SongBoom™ Boombox drop (e.g. gathered meat-clip)."""
    return get_property("_boomBoxFights", int)



from typing import Any

from jpype import JClass

from pymafia.ash.conversion import from_java, to_java
from pymafia.kolmafia import km

String = JClass("java.lang.String")
ByteArrayInputStream = JClass("java.io.ByteArrayInputStream")


class LibraryFunction:
    def __init__(self, name: str):
        self.name = name
        self.func = getattr(km.RuntimeLibrary, self.name)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name!r})"

    def __call__(self, *args, raw=False) -> Any:
        interpreter = km.AshRuntime()
        jargs = [to_java(arg) for arg in args]
        value = self.func(interpreter, *jargs)
        return value if raw else from_java(value)

    @property
    def signatures(self) -> list[str]:
        functions = km.RuntimeLibrary.getFunctions().findFunctions(self.name)
        return [f"{f.getType().toString()} {f.getSignature()}" for f in functions]


def script(lines: str, raw=False) -> Any:
    stream = ByteArrayInputStream(String(lines).getBytes())
    interpreter = km.AshRuntime()
    interpreter.validate(None, stream)
    value = interpreter.execute("main", None)
    return value if raw else from_java(value)


def ashref(command="") -> list[str]:
    names = set()
    for func in km.RuntimeLibrary.getFunctions():
        name = func.getName()
        if command.casefold() in name.casefold():
            names.add(name)
    return sorted(names)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.class_ import Class


@dataclass(frozen=True, order=True)
class Skill:
    id: int = km.DataTypes.SKILL_INIT.contentLong
    name: str = km.DataTypes.SKILL_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        id = km.SkillDatabase.getSkillId(key) if isinstance(key, str) else key
        name = km.SkillDatabase.getSkillName(id)
        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "id", id)
        object.__setattr__(self, "name", name)

    def __str__(self) -> str:
        ids = km.SkillDatabase.getSkillIds(self.name, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Skill]:
        from pymafia.ash import from_java

        values = km.DataTypes.SKILL_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def type(self) -> str:
        return km.SkillDatabase.getSkillTypeName(self.id)

    @property
    def level(self) -> int:
        return km.SkillDatabase.getSkillLevel(self.id)

    @property
    def image(self) -> str:
        return km.SkillDatabase.getSkillImage(self.id)

    @property
    def traincost(self) -> int:
        return km.SkillDatabase.getSkillPurchaseCost(self.id)

    @property
    def class_(self) -> Class:
        from pymafia.datatypes.class_ import Class

        try:
            return Class(km.SkillDatabase.getSkillCategory(self.id).name)
        except ValueError:
            return Class()

    @property
    def libram(self) -> bool:
        return km.SkillDatabase.isLibramSkill(self.id)

    @property
    def passive(self) -> bool:
        return km.SkillDatabase.isPassive(self.id)

    @property
    def buff(self) -> bool:
        return km.SkillDatabase.isBuff(self.id)

    @property
    def combat(self) -> bool:
        return km.SkillDatabase.isCombat(self.id)

    @property
    def song(self) -> bool:
        return km.SkillDatabase.isSong(self.id)

    @property
    def expression(self) -> bool:
        return km.SkillDatabase.isExpression(self.id)

    @property
    def walk(self) -> bool:
        return km.SkillDatabase.isWalk(self.id)

    @property
    def summon(self) -> bool:
        return km.SkillDatabase.isSummon(self.id)

    @property
    def permable(self) -> bool:
        return km.SkillDatabase.isPermable(self.id)

    @property
    def dailylimit(self) -> int:
        return km.SkillDatabase.getMaxCasts(self.id)

    @property
    def timescast(self) -> int:
        return km.SkillDatabase.getCasts(self.id)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.item import Item


@dataclass(frozen=True, order=True)
class Coinmaster:
    coinmaster: Any = field(default=km.DataTypes.COINMASTER_INIT.content, compare=False)
    name: str = km.DataTypes.COINMASTER_INIT.contentString

    def __init__(self, key: str | None = None):
        if (
            isinstance(key, str) and key.casefold() == self.name.casefold()
        ) or key is None:
            return

        coinmaster = km.CoinmasterRegistry.findCoinmaster(key)
        if coinmaster is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "coinmaster", coinmaster)
        object.__setattr__(self, "name", coinmaster.getMaster())

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Coinmaster]:
        from pymafia.ash import from_java

        values = km.DataTypes.COINMASTER_TYPE.allValues()
        return from_java(values)

    @property
    def token(self) -> str:
        return self.coinmaster.getToken() if self.coinmaster else ""

    @property
    def item(self) -> Item:
        from pymafia.datatypes.item import Item

        if self.coinmaster is None:
            return Item()
        item = self.coinmaster.getItem()
        return Item(item.getItemId()) if item is not None else Item()

    @property
    def preference(self) -> str:
        return self.coinmaster.getProperty() if self.coinmaster is not None else ""

    @property
    def available_tokens(self) -> int:
        return self.coinmaster.availableTokens() if self.coinmaster is not None else 0

    @property
    def buys(self) -> bool:
        return (
            self.coinmaster is not None and self.coinmaster.getSellAction() is not None
        )

    @property
    def sells(self) -> bool:
        return (
            self.coinmaster is not None and self.coinmaster.getBuyAction() is not None
        )

    @property
    def nickname(self) -> str:
        return self.coinmaster.getNickname() if self.coinmaster is not None else ""



import re

from pymafia import ash
from pymafia.preference import get_property


def have() -> bool:
    """Return True if the player has the shrine to the Barrel god unlocked, False otherwise."""
    return get_property("barrelShrineUnlocked", bool)


def smash_free() -> bool:
    """Smash the free barrels in The Barrel full of Barrels."""
    if not have():
        return False

    pattern = '<div class="ex"><a class="spot" href="([^"]+)"><img title="A barrel"'
    for barrel in re.findall(pattern, ash.visit_url("barrel.php")):
        ash.visit_url(barrel)
    return True



from enum import Enum, IntEnum

from pymafia import ash, player
from pymafia.combat import Macro
from pymafia.datatypes import Familiar, Item
from pymafia.preference import get_property

FAMILIAR = Familiar("God Lobster")


class Reward(IntEnum):
    REGALIA = 1
    BLESSING = 2
    EXPERIENCE = 3


class Regalia(Enum):
    SCEPTER = Item("God Lobster's Scepter")
    RING = Item("God Lobster's Ring")
    ROD = Item("God Lobster's Rod")
    ROBE = Item("God Lobster's Robe")
    CROWN = Item("God Lobster's Crown")


def have() -> bool:
    """Return True if the player has the God Lobster in their terrarium, False otherwise."""
    return player.have(FAMILIAR)


def fights_today() -> int:
    """Return the number of God Lobster fights used today."""
    return get_property("godLobsterFights", int)


def fights_left() -> int:
    """Return the number of God Lobster fights remaining today."""
    return 3 - fights_today()


def fight(reward: Reward, macro: Macro = Macro()) -> bool:
    """Fight the God Lobster and choose a reward."""
    if not have():
        return False
    if fights_left() < 1:
        return False

    ash.use_familiar(FAMILIAR)
    ash.visit_url("main.php?fightgodlobster=1")
    ash.run_combat(macro)
    ash.run_choice(int(reward))
    return True



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.stat import Stat


@dataclass(frozen=True, order=True)
class Class:
    ascension_class: Any = field(default=km.DataTypes.CLASS_INIT.content, compare=False)
    id: int = km.DataTypes.CLASS_INIT.contentLong
    name: str = km.DataTypes.CLASS_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        ascension_class = km.AscensionClass.find(key)
        if ascension_class is None or ascension_class.getId() < 0:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "ascension_class", ascension_class)
        object.__setattr__(self, "id", ascension_class.getId())
        object.__setattr__(self, "name", ascension_class.getName())

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Class]:
        from pymafia.ash import from_java

        values = km.DataTypes.CLASS_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def primestat(self) -> Stat:
        from pymafia.datatypes.stat import Stat

        if self.ascension_class is None:
            return Stat()
        prime_index = self.ascension_class.getPrimeStatIndex()
        name = km.AdventureResult.STAT_NAMES[prime_index]
        return Stat(name)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from pymafia.kolmafia import km


@dataclass(frozen=True, order=True)
class Servant:
    data: Any = field(default=km.DataTypes.SERVANT_INIT.content, compare=False)
    id: int = km.DataTypes.SERVANT_INIT.contentLong
    name: str = km.DataTypes.SERVANT_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        data = (
            km.EdServantData.typeToData(key)
            if isinstance(key, str)
            else km.EdServantData.idToData(key)
        )
        if data is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "data", data)
        object.__setattr__(self, "id", km.EdServantData.dataToId(data))
        object.__setattr__(self, "name", km.EdServantData.dataToType(data))

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Servant]:
        from pymafia.ash import from_java

        values = km.DataTypes.SERVANT_TYPE.allValues()
        return from_java(values)

    @property
    def servant(self) -> Any:
        return km.EdServantData.findEdServant(self.name)

    @property
    def level(self) -> int:
        return self.servant.getLevel() if self.servant is not None else 0

    @property
    def experience(self) -> int:
        return self.servant.getExperience() if self.servant is not None else 0

    @property
    def image(self) -> str:
        return km.EdServantData.dataToImage(self.data) if self.data is not None else ""

    @property
    def level1_ability(self) -> str:
        return (
            km.EdServantData.dataToLevel1Ability(self.data)
            if self.data is not None
            else ""
        )

    @property
    def level7_ability(self) -> str:
        return (
            km.EdServantData.dataToLevel7Ability(self.data)
            if self.data is not None
            else ""
        )

    @property
    def level14_ability(self) -> str:
        return (
            km.EdServantData.dataToLevel14Ability(self.data)
            if self.data is not None
            else ""
        )

    @property
    def level21_ability(self) -> str:
        return (
            km.EdServantData.dataToLevel21Ability(self.data)
            if self.data is not None
            else ""
        )



__all__ = ["LibraryFunction", "from_java", "to_java"]

from pymafia.ash.conversion import from_java, to_java
from pymafia.ash.library import LibraryFunction


def __getattr__(name: str) -> LibraryFunction:
    return LibraryFunction(name)



from enum import IntEnum

from pymafia import ash, player
from pymafia.datatypes import Item
from pymafia.preference import get_property

ITEM = Item("Fourth of May Cosplay Saber")


class Upgrade(IntEnum):
    NONE = 0
    MP = 1
    ML = 2
    RESISTANCE = 3
    FAMILIAR = 4


def have() -> bool:
    """Return True if the player has the Fourth of May Cosplay Saber available, False otherwise."""
    return player.have(ITEM)


def current_upgrade() -> Upgrade:
    """Return the current Fourth of May Cosplay Saber upgrade."""
    return Upgrade(get_property("_saberMod", int))


def is_upgraded() -> bool:
    """Return True if the Fourth of May Cosplay Saber has been upgraded today, False otherwise."""
    return current_upgrade() is not Upgrade.NONE


def upgrade(new_upgrade: Upgrade) -> bool:
    """Upgrade the Fourth of May Cosplay Saber."""
    if not have():
        return False
    if is_upgraded():
        return current_upgrade() == new_upgrade

    return ash.cli_execute(f"saber {new_upgrade.name.lower()}")



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from pymafia.kolmafia import km


@dataclass(frozen=True, order=True)
class Path:
    ascension_path: Any = field(default=km.DataTypes.PATH_INIT.content, compare=False)
    id: int = km.DataTypes.PATH_INIT.contentLong
    name: str = km.DataTypes.PATH_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        ascension_path = (
            km.AscensionPath.nameToPath(key)
            if isinstance(key, str)
            else km.AscensionPath.idToPath(key)
        )
        if ascension_path == km.AscensionPath.Path.NONE:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "ascension_path", ascension_path)
        object.__setattr__(self, "id", ascension_path.getId())
        object.__setattr__(self, "name", ascension_path.getName())

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Path]:
        from pymafia.ash import from_java

        values = km.DataTypes.PATH_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def avatar(self) -> bool:
        return self.ascension_path.isAvatar()

    @property
    def image(self) -> str:
        return self.ascension_path.getImage()

    @property
    def points(self) -> int:
        return self.ascension_path.getPoints()

    @property
    def familiars(self) -> bool:
        return self.ascension_path.canUseFamiliars()



from enum import Enum

from pymafia import ash, player
from pymafia.datatypes import Item, Location, Monster
from pymafia.preference import get_property

ITEM = Item("protonic accelerator pack")


class Ghost(Enum):
    OILY = Monster("the ghost of Oily McBindle")
    SKIN = Monster("boneless blobghost")
    BAGEL = Monster("the ghost of Monsieur Baguelle")
    HORSE = Monster("The Headless Horseman")
    ICEWOMAN = Monster("The Icewoman")
    SCREEGE = Monster("The ghost of Ebenoozer Screege")
    SRAVEN = Monster("The ghost of Lord Montague Spookyraven")
    HIPPY = Monster('the ghost of Vanillica "Trashblossom" Gorton')
    FLAME = Monster("the ghost of Sam McGee")
    CENSOR = Monster("The ghost of Richard Cockingham")
    WALDO = Monster("The ghost of Waldo the Carpathian")
    LIME = Monster("Emily Koops, a spooky lime")
    FORTUNADO = Monster("The ghost of Jim Unfortunato")


def have() -> bool:
    """Return True if the player has the protonic accelerator pack available, False otherwise."""
    return player.have(ITEM)


def ghost_location() -> Location:
    """Return the current location of the protonic ghost."""
    return get_property("ghostLocation", Location)


def streams_crossed() -> bool:
    """Return True if the player has crossed streams today, False otherwise."""
    return get_property("_streamsCrossed", bool)


def cross_streams() -> bool:
    """Cross streams with the target in preference "streamCrossDefaultTarget"."""
    if not have():
        return False
    if streams_crossed():
        return True

    return ash.cli_execute("crossstreams")



from pymafia import ash
from pymafia.preference import get_property

WANDERER_HOLIDAYS = {
    "El Dia De Los Muertos Borrachos",
    "Feast of Boris",
    "Talk Like a Pirate Day",
}


def holiday_wanderer_day() -> bool:
    """Return whether today is a holiday wanderer day."""
    holidays_today = set(ash.holiday().split("/"))
    return bool(holidays_today.intersection(WANDERER_HOLIDAYS))


def vote_wanderer_now() -> float:
    """Returns whether the player will encounter a vote wanderer on the next
    turn, providing an "I Voted!" sticker is equipped."""
    return (
        ash.total_turns_played() % 11 == 1
        and get_property("lastVoteMonsterTurn", int) < ash.total_turns_played()
    )


def kramco_wanderer_chance() -> float:
    """Returns the chance the player will encounter a sausage goblin on the
    next turn, providing the Kramco Sausage-o-Matic is equipped."""
    fights = get_property("_sausageFights", int)
    last_fight = get_property("_lastSausageMonsterTurn", int)
    total_turns = ash.total_turns_played()
    if fights < 1:
        return last_fight == total_turns and 0.5 if ash.my_turncount() < 1 else 1.0
    turns_since_last_fight = total_turns - last_fight
    return min(
        1.0, (turns_since_last_fight + 1) / (5 + fights * 3 + max(0, fights - 5) ** 3)
    )



__all__ = [
    "Bounty",
    "BountyType",
    "KoLInternalBountyType",
    "Class",
    "Coinmaster",
    "Effect",
    "EffectQuality",
    "Element",
    "Familiar",
    "Item",
    "CandyType",
    "Location",
    "Monster",
    "Path",
    "Phylum",
    "Servant",
    "Skill",
    "Slot",
    "Stat",
    "Thrall",
    "Vykea",
    "VykeaCompanionType",
    "VykeaRune",
]

from pymafia.datatypes.bounty import Bounty, BountyType, KoLInternalBountyType
from pymafia.datatypes.class_ import Class
from pymafia.datatypes.coinmaster import Coinmaster
from pymafia.datatypes.effect import Effect, EffectQuality
from pymafia.datatypes.element import Element
from pymafia.datatypes.familiar import Familiar
from pymafia.datatypes.item import CandyType, Item
from pymafia.datatypes.location import Location
from pymafia.datatypes.monster import Monster
from pymafia.datatypes.path import Path
from pymafia.datatypes.phylum import Phylum
from pymafia.datatypes.servant import Servant
from pymafia.datatypes.skill import Skill
from pymafia.datatypes.slot import Slot
from pymafia.datatypes.stat import Stat
from pymafia.datatypes.thrall import Thrall
from pymafia.datatypes.vykea import Vykea, VykeaCompanionType, VykeaRune

MAFIA_DATATYPES = (
    Item,
    Location,
    Class,
    Stat,
    Skill,
    Effect,
    Familiar,
    Slot,
    Monster,
    Element,
    Coinmaster,
    Phylum,
    Bounty,
    Thrall,
    Servant,
    Vykea,
    Path,
)



from html import escape

from jpype import JClass

from pymafia.kolmafia import km

ByteArrayOutputStream = JClass("java.io.ByteArrayOutputStream")
PrintStream = JClass("java.io.PrintStream")
OutputStream = JClass("java.io.OutputStream")


def launch_gui():
    """Launch the KoLmafia GUI."""
    km.KoLmafia.main(["--GUI"])


def login(username: str, password: str | None = None) -> bool:
    """Login to Kingdom of Loathing through KoLmafia."""
    if password is None:
        password = km.KoLmafia.getSaveState(username)

    request = km.LoginRequest(username, password)
    request.run()
    return request


def abort(message: str = ""):
    """Immediately halt KoLmafia."""
    km.KoLmafia.updateDisplay(km.KoLConstants.MafiaState.ABORT, message)


def log(message: str, html: bool = False):
    """Log a message in the KoLmafia CLI."""
    message = str(message)
    if not html:
        message = escape(message)

    km.RequestLogger.printLine(message)


def execute(command: str) -> str:
    """Execute a command in the KoLmafia CLI and return the output."""
    ostream = OutputStream @ ByteArrayOutputStream()
    out = PrintStream(ostream)
    km.RequestLogger.openCustom(out)
    km.KoLmafiaCLI.DEFAULT_SHELL.executeLine(command)
    return ostream.toString()



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum

from pymafia.kolmafia import km


class EffectQuality(IntEnum):
    NONE = -1
    GOOD = 0
    NEUTRAL = 1
    BAD = 2


@dataclass(frozen=True, order=True)
class Effect:
    id: int = km.DataTypes.EFFECT_INIT.contentLong
    name: str = km.DataTypes.EFFECT_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.name.casefold()) or key in (
            self.id,
            None,
        ):
            return

        id = km.EffectDatabase.getEffectId(key) if isinstance(key, str) else key
        name = km.EffectDatabase.getEffectName(id)
        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "id", id)
        object.__setattr__(self, "name", name)

    def __str__(self) -> str:
        ids = km.EffectDatabase.getEffectIds(self.name, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Effect]:
        from pymafia.ash import from_java

        values = km.DataTypes.EFFECT_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def default(self) -> str:
        return km.EffectDatabase.getDefaultAction(self.id) or ""

    @property
    def actions(self) -> list[str]:
        return list(km.EffectDatabase.getAllActions(self.id))

    @property
    def quality(self) -> EffectQuality:
        return EffectQuality(km.EffectDatabase.getQuality(self.id))

    @property
    def attributes(self) -> list[str]:
        attrs = km.EffectDatabase.getEffectAttributes(self.id)
        return list(attrs) if attrs is not None else []

    @property
    def note(self) -> str:
        return km.EffectDatabase.getActionNote(self.id) or ""

    @property
    def image(self) -> str:
        return km.EffectDatabase.getImageName(self.id)

    @property
    def descid(self) -> str:
        return km.EffectDatabase.getDescriptionId(self.id) or ""

    @property
    def candy_tier(self) -> int:
        return km.CandyDatabase.getEffectTier(self.id)

    @property
    def song(self) -> bool:
        return km.EffectDatabase.isSong(self.id)



from collections import abc
from typing import Any

from jpype import JClass

from pymafia.datatypes import (
    MAFIA_DATATYPES,
    Bounty,
    Class,
    Coinmaster,
    Effect,
    Element,
    Familiar,
    Item,
    Location,
    Monster,
    Path,
    Phylum,
    Servant,
    Skill,
    Slot,
    Stat,
    Thrall,
    Vykea,
)
from pymafia.kolmafia import km

TreeMap = JClass("java.util.TreeMap")
ArrayList = JClass("java.util.ArrayList")

TYPE_CONVERSIONS = {
    km.DataTypes.BOOLEAN_TYPE: bool,
    km.DataTypes.INT_TYPE: int,
    km.DataTypes.FLOAT_TYPE: float,
    km.DataTypes.STRING_TYPE: str,
    km.DataTypes.BUFFER_TYPE: str,
    km.DataTypes.ITEM_TYPE: Item,
    km.DataTypes.LOCATION_TYPE: Location,
    km.DataTypes.CLASS_TYPE: Class,
    km.DataTypes.STAT_TYPE: Stat,
    km.DataTypes.SKILL_TYPE: Skill,
    km.DataTypes.EFFECT_TYPE: Effect,
    km.DataTypes.FAMILIAR_TYPE: Familiar,
    km.DataTypes.SLOT_TYPE: Slot,
    km.DataTypes.MONSTER_TYPE: Monster,
    km.DataTypes.ELEMENT_TYPE: Element,
    km.DataTypes.COINMASTER_TYPE: Coinmaster,
    km.DataTypes.PHYLUM_TYPE: Phylum,
    km.DataTypes.BOUNTY_TYPE: Bounty,
    km.DataTypes.THRALL_TYPE: Thrall,
    km.DataTypes.SERVANT_TYPE: Servant,
    km.DataTypes.VYKEA_TYPE: Vykea,
    km.DataTypes.PATH_TYPE: Path,
}


def to_java(obj: Any) -> Any:
    """Convert a Python object to a KoLmafia Value or PluralValue."""
    if isinstance(obj, (bool, int, float, str)):
        return km.Value(obj)

    if isinstance(obj, MAFIA_DATATYPES):
        parser = getattr(km.DataTypes, f"parse{type(obj).__name__}Value")
        return parser(str(obj), False)

    if isinstance(obj, abc.Mapping):
        jmap = TreeMap()
        for k, v in obj.items():
            jk = to_java(k)
            jv = to_java(v)
            jmap.put(jk, jv)
        data_type = jmap.firstEntry().getValue().getType()
        index_type = jmap.firstEntry().getKey().getType()
        aggregate_type = km.AggregateType(data_type, index_type)
        return km.MapValue(aggregate_type, jmap)

    if isinstance(obj, abc.Iterable):
        jlist = ArrayList()
        for item in obj:
            jitem = to_java(item)
            jlist.add(jitem)
        data_type = jlist.get(0).getType()
        size = jlist.size()
        aggregate_type = km.AggregateType(data_type, size)
        return km.ArrayValue(aggregate_type, jlist)

    raise TypeError(f"unsupported type: {type(obj).__name__!r}")


def from_java(obj: Any) -> Any:
    """Convert a KoLmafia Value or PluralValue to a Python object."""
    jtype = obj.getType()

    if jtype == km.DataTypes.VOID_TYPE:
        return None

    if jtype in TYPE_CONVERSIONS:
        return TYPE_CONVERSIONS[jtype](obj.toJSON())

    if isinstance(jtype, km.AggregateType) and isinstance(obj.content, abc.Mapping):
        return {from_java(k): from_java(v) for k, v in obj.content.items()}

    if isinstance(jtype, km.AggregateType) and isinstance(obj.content, abc.Iterable):
        return [from_java(x) for x in obj.content]

    raise TypeError(f"unsupported type: {jtype.getName()!r}")



from enum import IntEnum
from typing import Tuple

from pymafia import ash, player
from pymafia.datatypes import Familiar, Item
from pymafia.preference import get_property

FAMILIAR = Familiar("Crimbo Shrub")
DECORATIONS = Item("box of old Crimbo decorations")


class Topper(IntEnum):
    MUSCLE = 1
    MYSTICALITY = 2
    MOXIE = 3


class Lights(IntEnum):
    PRISMATIC = 1
    HOT = 2
    COLD = 3
    STENCH = 4
    SPOOKY = 5
    SLEAZE = 6


class Garland(IntEnum):
    HP = 1
    PVP = 2
    BLOCKING = 3


class Gift(IntEnum):
    YELLOW = 1
    MEAT = 2
    GIFTS = 3


def have() -> bool:
    """Return True is the player has the Crimbo Shrub in their terrarium, False otherwise."""
    return player.have(FAMILIAR)


def is_decorated() -> bool:
    """Return True if the Crimbo Shrub has been decorated today, False otherwise."""
    return get_property("_shrubDecorated", bool)


def current_decorations() -> Tuple[Topper, Lights, Garland, Gift]:
    """Current Crimbo Shrub decorations."""
    return (
        Topper(get_property("shrubTopper", int)),
        Lights(get_property("shrubLights", int)),
        Garland(get_property("shrubGarland", int)),
        Gift(get_property("shrubGifts", int)),
    )


def decorate(topper: Topper, lights: Lights, garland: Garland, gift: Gift) -> bool:
    """Decorate the Crimbo Shrub."""
    if not have():
        return False
    if is_decorated():
        return current_decorations() == (topper, lights, garland, gift)

    if not player.have(DECORATIONS):
        ash.use_familiar(FAMILIAR)
    ash.visit_url(f"inv_use.php?pwd=&which=99&whichitem={DECORATIONS.id}")
    ash.visit_url(
        f"choice.php?whichchoice=999&pwd=&option=1&topper={topper}&lights={lights}&garland={garland}&gift={gift}"
    )
    return True



import re

from pymafia import ash
from pymafia.datatypes import Effect, Familiar, Item, Monster, Servant, Skill


def have(thing: Effect | Familiar | Item | Servant | Skill, quantity: int = 1) -> bool:
    """Return whether the player "has" any entity which one could feasibly "have"."""
    if isinstance(thing, Effect):
        return ash.have_effect(thing) >= quantity
    if isinstance(thing, Familiar):
        return ash.have_familiar(thing)
    if isinstance(thing, Item):
        return ash.available_amount(thing) >= quantity
    if isinstance(thing, Servant):
        return ash.have_servant(thing)
    if isinstance(thing, Skill):
        return ash.have_skill(thing)
    raise TypeError(f"unexpected type {type(thing).__name__!r}")


def in_choice(choice: int | str) -> bool:
    """Return whether the player is handling a particular choice adventure."""
    return ash.handling_choice() and ash.last_choice() == int(choice)


def in_combat(monster: Monster | None = None) -> bool:
    """Return where the player is in combat with a particular monster."""
    if ash.current_round() < 1:
        return False
    if monster is None:
        return True

    page = ash.visit_url("fight.php")
    match = re.search("<!-- MONSTERID: (\\d+) -->", page)
    if not match:
        raise RuntimeError("unable to identify monster")
    return int(match.group(1)) == monster.id


def can_visit_url() -> bool:
    """Return whether the player can safely visit an in-game url."""
    return not (
        ash.current_round() > 0
        or ash.in_multi_fight()
        or ash.fight_follows_choice()
        or ash.handling_choice()
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.skill import Skill


@dataclass(frozen=True, order=True)
class Thrall:
    data: Any = field(default=km.DataTypes.THRALL_INIT.content, compare=False)
    id: int = km.DataTypes.THRALL_INIT.contentLong
    type: str = km.DataTypes.THRALL_INIT.contentString

    def __init__(self, key: int | str | None = None):
        if (isinstance(key, str) and key.casefold() == self.type.casefold()) or key in (
            self.id,
            None,
        ):
            return

        data = (
            km.PastaThrallData.typeToData(key)
            if isinstance(key, str)
            else km.PastaThrallData.idToData(key)
        )
        if data is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "data", data)
        object.__setattr__(self, "id", km.PastaThrallData.dataToId(data))
        object.__setattr__(self, "type", km.PastaThrallData.dataToType(data))

    def __str__(self) -> str:
        return self.type

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Thrall]:
        from pymafia.ash import from_java

        values = km.DataTypes.THRALL_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def thrall(self) -> Any:
        return km.KoLCharacter.findPastaThrall(self.type)

    @property
    def name(self) -> str:
        return self.thrall.getName() if self.thrall else ""

    @property
    def level(self) -> int:
        return self.thrall.getLevel() if self.thrall else 0

    @property
    def image(self) -> str:
        return (
            km.PastaThrallData.dataToImage(self.data) if self.data is not None else ""
        )

    @property
    def tinyimage(self) -> str:
        return (
            km.PastaThrallData.dataToTinyImage(self.data)
            if self.data is not None
            else ""
        )

    @property
    def skill(self) -> Skill:
        from pymafia.datatypes.skill import Skill

        return Skill(
            km.PastaThrallData.dataToSkillId(self.data)
            if self.data is not None
            else None
        )

    @property
    def current_modifiers(self) -> str:
        return self.thrall.getCurrentModifiers() if self.thrall else ""



__version__ = "0.0.0"  # This is the standard placeholder for poetry-dynamic-versioning

from pymafia import ash
from pymafia.ash import *
from pymafia.combat import Macro
from pymafia.datatypes import *
from pymafia.iotms import *
from pymafia.kolmafia import *
from pymafia.player import *
from pymafia.preference import *
from pymafia.utils import *
from pymafia.wanderer import *



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.location import Location
    from pymafia.datatypes.monster import Monster


class BountyType(Enum):
    NONE = None
    EASY = "easy"
    HARD = "hard"
    SPECIAL = "special"


class KoLInternalBountyType(Enum):
    NONE = BountyType.NONE
    LOW = BountyType.EASY
    HIGH = BountyType.HARD
    SPECIAL = BountyType.SPECIAL


@dataclass(frozen=True, order=True)
class Bounty:
    name: str = km.DataTypes.BOUNTY_INIT.contentString

    def __init__(self, key: str | None = None):
        if (
            isinstance(key, str) and key.casefold() == self.name.casefold()
        ) or key is None:
            return

        bounties = km.BountyDatabase.getMatchingNames(key)
        if len(bounties) != 1:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        canonical = bounties[0]
        object.__setattr__(self, "name", km.BountyDatabase.canonicalToName(canonical))

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Bounty]:
        from pymafia.ash import from_java

        values = km.DataTypes.BOUNTY_TYPE.allValues()
        return from_java(values)

    @property
    def plural(self) -> str:
        return km.BountyDatabase.getPlural(self.name) or ""

    @property
    def type(self) -> BountyType:
        return BountyType(km.BountyDatabase.getType(self.name))

    @property
    def kol_internal_type(self) -> KoLInternalBountyType:
        return KoLInternalBountyType(self.type)

    @property
    def number(self) -> int:
        return km.BountyDatabase.getNumber(self.name)

    @property
    def image(self) -> str:
        return km.BountyDatabase.getImage(self.name) or ""

    @property
    def monster(self) -> Monster:
        from pymafia.datatypes.monster import Monster

        return Monster(km.BountyDatabase.getMonster(self.name))

    @property
    def location(self) -> Location:
        from pymafia.datatypes.location import Location

        return Location(km.BountyDatabase.getLocation(self.name))



from typing import Any, Type, TypeVar, overload

from pymafia.kolmafia import km

T = TypeVar("T")


@overload
def get_property(name: str) -> str:
    ...


@overload
def get_property(name: str, t: Type[T]) -> T:
    ...


def get_property(name, t=str):
    """Return the value of a KoLmafia property."""
    if t is bool:
        return km.Preferences.getBoolean(name)
    if t is int:
        return km.Preferences.getInteger(name)
    if t is float:
        return km.Preferences.getFloat(name)
    return t(km.Preferences.getString(name))


def set_property(name: str, value: Any = ""):
    """Set the value of a KoLmafia property."""
    if isinstance(value, bool):
        km.Preferences.setBoolean(name, value)
    elif isinstance(value, int):
        km.Preferences.setInteger(name, value)
    elif isinstance(value, float):
        km.Preferences.setFloat(name, value)
    else:
        km.Preferences.setString(name, str(value))


def set_choice(choice: int, value: int | str):
    """Set the value of a KoLmafia choice adventure property."""
    set_property(f"choiceAdventure{choice}", value)



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any

from pymafia.kolmafia import km

if TYPE_CHECKING:
    from pymafia.datatypes.element import Element


class VykeaCompanionType(Enum):
    NONE = km.VYKEACompanionData.VYKEACompanionType.NONE
    BOOKSHELF = km.VYKEACompanionData.VYKEACompanionType.BOOKSHELF
    DRESSER = km.VYKEACompanionData.VYKEACompanionType.DRESSER
    CEILING_FAN = km.VYKEACompanionData.VYKEACompanionType.CEILING_FAN
    COUCH = km.VYKEACompanionData.VYKEACompanionType.COUCH
    LAMP = km.VYKEACompanionData.VYKEACompanionType.LAMP
    DISHRACK = km.VYKEACompanionData.VYKEACompanionType.DISHRACK


class VykeaRune(Enum):
    NONE = km.VYKEACompanionData.NO_RUNE
    FRENZY = km.VYKEACompanionData.FRENZY_RUNE
    BLOOD = km.VYKEACompanionData.BLOOD_RUNE
    LIGHTNING = km.VYKEACompanionData.LIGHTNING_RUNE


@dataclass(frozen=True, order=True)
class Vykea:
    companion: Any = field(default=km.VYKEACompanionData.NO_COMPANION, compare=False)
    type: VykeaCompanionType = VykeaCompanionType(companion.default.getType())
    rune: VykeaRune = VykeaRune(companion.default.getRune())
    level: int = companion.default.getLevel()

    def __init__(self, key: str | None = None):
        if (
            isinstance(key, str)
            and key.casefold() == km.DataTypes.VYKEA_INIT.contentString.casefold()
        ) or key is None:
            return

        companion = km.VYKEACompanionData.fromString(key)
        if companion is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "companion", companion)
        object.__setattr__(self, "type", VykeaCompanionType(companion.getType()))
        object.__setattr__(self, "rune", VykeaRune(companion.getRune()))
        object.__setattr__(self, "level", companion.getLevel())

    def __str__(self) -> str:
        return (
            self.companion.toString() if self else km.DataTypes.VYKEA_INIT.contentString
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Vykea]:
        from pymafia.ash import from_java

        values = km.DataTypes.VYKEA_TYPE.allValues()
        return from_java(values)

    @property
    def name(self) -> str:
        return self.companion.getName()

    @property
    def image(self) -> str:
        return self.companion.getImage()

    @property
    def modifiers(self) -> str:
        return self.companion.getModifiers()

    @property
    def attack_element(self) -> Element:
        from pymafia.datatypes.element import Element

        return Element(self.companion.getAttackElement().toString())



from enum import IntEnum

from pymafia import ash
from pymafia.combat import Macro
from pymafia.datatypes import Location
from pymafia.preference import get_property, set_property

LOCATION = Location("The Tunnel of L.O.V.E.")


class LOVEquipment(IntEnum):
    EARDIGAN = 1
    EPAULETTES = 2
    EARRING = 3


class LOVEffect(IntEnum):
    LOVEBOTAMY = 1
    OPEN_HEART = 2
    WANDERING_EYE = 3


class LOVItem(IntEnum):
    ENAMORANG = 1
    EMOTIONIZER = 2
    CHOCOLATE = 3
    BOUQUET = 4
    ELEPHANT = 5
    TOAST = 6
    NONE = 7


def have() -> bool:
    """Return True if the player has The Tunnel of L.O.V.E available, False otherwise."""
    return get_property("loveTunnelAvailable", bool)


def is_used() -> bool:
    """Return True if The Tunnel of L.O.V.E has been used today, False otherwise."""
    return get_property("_loveTunnelUsed", bool)


def fight_all(
    equipment: LOVEquipment, effect: LOVEffect, item: LOVItem, macro: Macro = Macro()
) -> bool:
    """Fight all LOV monsters and get buffs/equipment."""
    if not have():
        return False
    if is_used():
        return True

    set_property("choiceAdventure1222", 1)  # Entrance
    set_property("choiceAdventure1223", 1)  # Fight LOV Enforcer
    set_property("choiceAdventure1224", int(equipment))
    set_property("choiceAdventure1225", 1)  # Fight LOV Engineer
    set_property("choiceAdventure1226", int(effect))
    set_property("choiceAdventure1227", 1)  # Fight LOV Equivocator
    set_property("choiceAdventure1228", int(item))
    return ash.adv1(LOCATION, -1, macro)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from pymafia.kolmafia import km


@dataclass(frozen=True, order=True)
class Phylum:
    phylum: Any = field(default=km.DataTypes.PHYLUM_INIT.content, compare=False)
    name: str = km.DataTypes.PHYLUM_INIT.contentString

    def __init__(self, key: str | None = None):
        if (
            isinstance(key, str) and key.casefold() == self.name.casefold()
        ) or key is None:
            return

        phylum = km.MonsterDatabase.Phylum.find(key)
        if phylum == km.MonsterDatabase.Phylum.NONE:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        object.__setattr__(self, "phylum", phylum)
        object.__setattr__(self, "name", phylum.toString())

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self)!r})"

    def __bool__(self) -> bool:
        return self != type(self)()

    @classmethod
    def all(cls) -> list[Phylum]:
        from pymafia.ash import from_java

        values = km.DataTypes.PHYLUM_TYPE.allValues()
        return sorted(from_java(values))

    @property
    def image(self) -> str:
        if self.phylum == km.MonsterDatabase.Phylum.NONE:
            return ""
        return self.phylum.getImage()



from enum import IntEnum

from pymafia import ash, player
from pymafia.datatypes import Element, Item, Stat

ITEM = Item("portable pantogram")
PANTS = Item("pantogram pants")


class PantogramAlignment(IntEnum):
    MUSCLE = 1
    MYSTICALITY = 2
    MOXIE = 3


class PantogramElement(IntEnum):
    HOT = 1
    COLD = 2
    SPOOKY = 3
    SLEAZE = 4
    STENCH = 5


alignments = {Stat("Muscle"): 1, Stat("Mysticality"): 2, Stat("Moxie"): 3}
elements = {
    Element("hot"): 1,
    Element("cold"): 2,
    Element("spooky"): 3,
    Element("sleaze"): 4,
    Element("stench"): 5,
}
left_sacrifices = {
    "Maximum HP: 40": (-1, 0),
    "Maximum MP: 20": (-2, 0),
    "HP Regen Max: 10": (Item("red pixel potion"), 1),
    "HP Regen Max: 15": (Item("royal jelly"), 1),
    "HP Regen Max: 20": (Item("scented massage oil"), 1),
    "MP Regen Max: 10": (Item("Cherry Cloaca Cola"), 1),
    "MP Regen Max: 15": (Item("bubblin' crude"), 1),
    "MP Regen Max: 20": (Item("glowing New Age crystal"), 1),
    "Mana Cost: -3": (Item("baconstone"), 1),
}
right_sacrifices = {
    "Weapon Damage: 20": (-1, 0),
    "Spell Damage Percent: 20": (-2, 0),
    "Meat Drop: 30": (Item("taco shell"), 1),
    "Meat Drop: 60": (Item("porquoise"), 1),
    "Item Drop: 15": (Item("fairy gravy boat"), 1),
    "Item Drop: 30": (Item("tiny dancer"), 1),
    "Muscle Experience: 3": (Item("Knob Goblin firecracker"), 3),
    "Mysticality Experience: 3": (Item("razor-sharp can lid"), 3),
    "Moxie Experience: 3": (Item("spider web"), 3),
    "Muscle Experience Percent: 25": (Item("synthetic marrow"), 5),
    "Mysticality Experience Percent: 25": (Item("haunted battery"), 5),
    "Moxie Experience Percent: 25": (Item("the funk"), 5),
}
middle_sacrifices = {
    "Combat Rate: -5": (-1, 0),
    "Combat Rate: 5": (-2, 0),
    "Critical Hit Percent: 10": (Item("hamethyst"), 1),
    "Initiative: 50": (Item("bar skin"), 1),
    "Familiar Weight: 10": (Item("lead necklace"), 11),
    "Candy Drop: 100": (Item("huge bowl of candy"), 1),
    "Item Drop Penalty: -10": (Item("sea salt crystal"), 11),
    "Fishing Skill: 5": (Item("wriggling worm"), 1),
    "Pool Skill: 5": (Item("8-ball"), 15),
    "Avatar: Purple": (Item("moxie weed"), 99),
    "Drops Items: true": (Item("ten-leaf clover"), 1),
}


def have():
    """Return True if the player has the portable pantogram available, False otherwise."""
    return player.have(ITEM)


def have_pants():
    """Return True if the player has the pantogram pants available today, False otherwise."""
    return player.have(PANTS)


def summon_pants(alignment, element, left, right, middle):
    """Summon pantogram pants."""
    if not have():
        return False
    if have_pants():
        return False  # TODO check enchantments

    m = alignments[alignment]
    e = elements[element]

    sacrifice, quantity = left_sacrifices[left]
    if isinstance(sacrifice, Item) and not player.have(sacrifice, quantity):
        return False
    s1 = f"{int(sacrifice)},{quantity}"

    sacrifice, quantity = right_sacrifices[right]
    if isinstance(sacrifice, Item) and not player.have(sacrifice, quantity):
        return False
    s2 = f"{int(sacrifice)},{quantity}"

    sacrifice, quantity = middle_sacrifices[middle]
    if isinstance(sacrifice, Item) and not player.have(sacrifice, quantity):
        return False
    s3 = f"{int(sacrifice)},{quantity}"

    ash.visit_url(
        f"choice.php?pwd=&whichchoice=1270&option=1&m={m}&e={e}&s1={s1}&s2={s2}&s3={s3}"
    )
    return True



import json
import os
import re
import urllib.request
import zipfile
from typing import Any, Iterable

import jpype

from pymafia.kolmafia import patch

JENKINS_JOB_URL = "https://ci.kolmafia.us/job/Kolmafia/lastSuccessfulBuild/"
JAR_LOCATION = "./kolmafia.jar"
JAVA_PATTERN = "(net\\/sourceforge\\/kolmafia.*\\/([^\\$]*))\\.class"


def download_kolmafia(location: str):
    with urllib.request.urlopen(JENKINS_JOB_URL + "/api/json") as response:
        data = json.loads(response.read().decode())
        jar_url = JENKINS_JOB_URL + "artifact/" + data["artifacts"][0]["relativePath"]
        urllib.request.urlretrieve(jar_url, filename=location)


class KoLmafia:
    def __init__(self, location: str):
        if not os.path.isfile(location):
            download_kolmafia(location)

        jpype.startJVM(classpath=location, convertStrings=True)
        patch.apply()

        self._classes = {}
        with zipfile.ZipFile(location) as archive:
            for filename in archive.namelist():
                if match := re.search(JAVA_PATTERN, filename):
                    self._classes[match.group(2)] = match.group(1)

    def __dir__(self) -> Iterable[str]:
        return list(self._classes.keys())

    def __getattr__(self, name: str) -> Any:
        if name in self._classes:
            return jpype.JClass(self._classes[name])
        return super().__getattribute__(name)


km = KoLmafia(JAR_LOCATION)



import _jpype
import jpype
import wrapt

PATCH_ENABLED = True


class KoLmafiaError(Exception):
    pass


# wrapt's wrap_function_wrapper does not allow specifying "enabled"
def wrap_function_wrapper(module, name, wrapper, enabled=None):
    return wrapt.wrap_object(
        module, name, wrapt.FunctionWrapper, (wrapper,), {"enabled": enabled}
    )


def apply():
    """Patch the JPype module to intercept all Java method calls.

    This patch does the following:
    * Automatically cast method arguments of type int to java.lang.Integer
    * Raise an exception if mafia is in a non-continue state after method invocation
    """
    KoLmafia = jpype.JClass("net.sourceforge.kolmafia.KoLmafia")

    def wrapper(wrapped, instance, args, kwargs):
        global PATCH_ENABLED
        try:
            PATCH_ENABLED = False

            args = [
                jpype.JInt(arg)
                if isinstance(arg, int) and not isinstance(arg, bool)
                else arg
                for arg in args
            ]
            result = wrapped(*args, **kwargs)

            if not KoLmafia.permitsContinue():
                KoLmafia.forceContinue()
                raise KoLmafiaError(KoLmafia.getLastMessage())

            return result
        finally:
            PATCH_ENABLED = True

    wrap_function_wrapper(
        _jpype, "_JMethod.__call__", wrapper, enabled=lambda: PATCH_ENABLED
    )
    wrap_function_wrapper(
        _jpype, "_JClass.__call__", wrapper, enabled=lambda: PATCH_ENABLED
    )
